import re
from typing import Dict, Any

class BurmeseConverter:
    def __init__(self):
        self.font_dictionary = {
                    '္': '်', '်': 'ျ', 'ျ': 'ြ', 'ြ': 'ွ', 'ွ': 'ှ', 'ႆ': 'ဿ', 'ဳ': 'ု', 'ဴ': 'ူ', 'ဿ': 'ူ',
                    '၀': 'ဝ', '၎': '၎င်း', 'ၚ': 'ါ်', 'ၠ': '္က', 'ၡ': '္ခ', 'ၢ': '္ဂ', 'ၣ': '္ဃ', 'ၥ': '္စ',
                    'ၦ': '္ဆ', 'ၧ': '္ဆ', 'ၨ': '္ဇ', 'ၩ': '္ဈ', 'ၪ': 'ဉ', 'ၫ': 'ည', 'ၬ': '္ဋ', 'ၭ': '္ဌ',
                    'ၮ': 'ဍ္ဍ', 'ၯ': 'ဍ္ဎ', 'ၰ': '္ဏ', 'ၱ': '္တ', 'ၲ': '္တ', 'ၳ': '္ထ', 'ၴ': '္ထ', 'ၵ': '္ဒ',
                    'ၶ': '္ဓ', 'ၷ': '္န', 'ၸ': '္ပ', 'ၹ': '္ဖ', 'ၺ': '္ဗ', 'ၻ': '္ဘ', 'ၼ': '္မ', 'ၽ': 'ျ',
                    'ၾ': 'ြ', 'ၿ': 'ြ', 'ႀ': 'ြ', 'ႁ': 'ြ', 'ႂ': 'ြ', 'ႃ': 'ြ', 'ႄ': 'ြ', 'ႅ': '္လ', 'ႇ': 'ှ',
                    'ႈ': 'ှု', 'ႉ': 'ှူ', 'ႊ': 'ွှ', 'ႎ': 'ိံ', 'ႏ': 'န', '႐': 'ရ', '႑': 'ဏ္ဍ', '႒': 'ဋ္ဌ',
                    '႓': '္ဘ', '႔': '့', '႕': '့', '႖': '္တွ', '႗': 'ဋ္ဋ', 'ၤ': 'င်္'
                }

    def zawgyi_to_unicode(self, text):
        # Replace characters based on font dictionary
        char_list = list(text)
        for i, char in enumerate(char_list):
            if char in self.font_dictionary:
                char_list[i] = self.font_dictionary[char]

        # Join modified characters back into a string
        converted_text = "".join(char_list)

        # Regular expression pattern for ordering characters in Unicode
        uni_pattern = r"(ေ)?(ြ)?([က-ဩဿၐ-ၕၚ-ၝၡၥၦၮ-ၰၵ-ႁႎ႐-႙႞႟])(ွ)?(ှ)?(ျ)?(င်္)?(ွ)?(ှ)?(ာ)?(း)?(္[က-ဘမလ])?"

        # Reorder characters to match Unicode standard
        result = re.sub(uni_pattern, r"\7\3\12\2\6\8\4\9\5\1\10\11", converted_text)

        return result

    def syllable_tokenization(self, option: int, user_input: str) -> str:
        """
        Process Burmese text by tokenizing and normalizing it.
        
        :param option: Processing option, 1 to include virama mark, others without
        :param user_input: Text to be processed
        :return: Processed text with or without virama mark
        """
        try:
            if option == 1:
                processed_input = re.sub(r"([က-အ|ဥ|ဦ](င်္|[က-အ][ှ]*[့း]*[်]|([က-အ]္)|[ါ-ှႏꩻ][ꩻ]*){0,}|.)", r"\1 ", user_input)
                processed_input = re.sub(r"(([က-အ])္ ([က-အ]))", r"\2် \3", processed_input)
                return f"With the virama mark: {processed_input}"
            else:
                processed_input = re.sub(r"([က-အ|ဥ|ဦ](င်္|[က-အ][ှ]*[့း]*[်]|([က-အ]္)|[ါ-ှႏꩻ][ꩻ]*){0,}|.)", r"\1 ", user_input)
                return f"Without the virama mark: {processed_input}"
        except Exception as e:
            raise Exception(f"An error occurred during text processing: {e}")

    def burmese_to_romanization(self, text: str) -> str:
        """
        Convert Burmese text to Romanized form.
        
        :param text: Burmese text input
        :return: Romanized text output
        """
        burmese_to_roman: Dict[str, str] = {
        'က': 'k', 'ခ': 'K', 'ဂ': 'g', 'ဃ': 'G', 'င': 'c', "၏": "E", "၍": "rx", "၌": "Nx", "င်္": "f",
        'စ': 's', 'ဆ': 'S', 'ဇ': 'z', 'ဈ': 'Z', "ဉ": "q", 'ည': 'Q', "ဋ": "tx", "ဌ": "Tx", "ဍ": "dx", "ဎ": "Dx", "ဏ": "nx",
        "ရ": "r", "ဓ": "D", "တ": "t", "ထ": "T", "ဒ": "d", "န": "n", "ပ": "p", "ဖ": "P", "ဗ": "b", "ဘ": "B", "မ": "m",
        "ယ": "y", "ဝ": "w", "သ": "j", "ဟ": "h", "အ": "a", 'လ': 'l', "ဠ": "lx", "ဣ": "ix", "ဤ": "Ix", "၊": "/", "။": "//", "ဥ": "Ux", "ဦ": "OO", "ဧ": "ax", "ဩ": "O", "ဪ": "OR", "ါ": "A", "ာ": "A", "ိ": "i", "ီ": "I","ေ": "e",
        "ု": "u", "ူ": "U", "ဲ": "L", "ံ": "N", "့": ".", "း": ":", "ျ": "Y", "ြ": "R", "ွ": "W", "ှ": "H","၎":"4",
        "ဿ": "jx", "်": ""
    }

        try:
    # Preserve text inside braces by identifying it with regex
            segments = re.split(r'({.*?})', text)  # Split the text by braces

            for i, segment in enumerate(segments):
                # Skip romanization for text inside braces
                if segment.startswith('{') and segment.endswith('}'):
                    continue
                tokenize_text = re.sub(r"([က-အ|ဥ|ဦ](င်္|[က-အ][ှ]*[့း]*[်]|([က-အ]္)|[ါ-ှႏꩻ][ꩻ]*){0,}|.)", r"\1 ", segment.strip())
                white_space_replace_comma = re.sub(r"[\s]{1}", r",", tokenize_text) # 3 line paung p try
                comma_replace_white_space = re.sub(r"[,]{3}",r" ",white_space_replace_comma) # 3 line paung p try
                remove_comma = re.sub(r",$",r'',comma_replace_white_space) # 3 line paung p try
                add_a_thet_behind_pat_sint = re.sub(r"(([က-အ])္ ([က-အ]))", r"\2် \3", remove_comma)

                # List of custom rules for converting specific word patterns to Romanized forms
                rules = [
                    (re.compile(r'ကျွန် မ '), "q'm "),
                    (re.compile(r'ကျွန် တော် '), "q't "),
                    (re.compile(r'ကျွန်ုပ် '), 'Q" '),
                ]
                for rule in rules:
                    text = rule[0].sub(rule[1], add_a_thet_behind_pat_sint)
                text = re.sub(r'([‌ေ][က-ဪ]*[ာါ]*[်])', r'\1F', text)

                # Perform Romanization by replacing Burmese characters with Roman equivalents
                for burmese_char, roman_char in sorted(burmese_to_roman.items(), key=lambda x: len(x[0]), reverse=True):
                    text = text.replace(burmese_char, roman_char)

                text = re.sub(r' ,', "", text)
                segments[i] = re.sub(r"္,(.)", r"\1,", text)
            result = ' '.join(segments)
            result = re.sub(r'\s+{', r' {', result)
            result = re.sub(r'}\s+', r'} ', result)
            return result
        except Exception as e:
            raise Exception(f"An error occurred during Burmese_to_Romanization: {e}")

    def romanization_to_burmese(self,burmese_text):
    
            burmese_to_roman = {
            'က': 'k', 'ခ': 'K', 'ဂ': 'g', 'ဃ': 'G', 'င': 'c', "၏": "E", "၍": "rx", "၌": "Nx", "င်္": "f",
            'စ': 's', 'ဆ': 'S', 'ဇ': 'z', 'ဈ': 'Z', "ဉ": "q", 'ည': 'Q', "ဋ": "tx", "ဌ": "Tx", "ဍ": "dx", "ဎ": "Dx", "ဏ": "nx",
            "ရ": "r", "ဓ": "D", "တ": "t", "ထ": "T", "ဒ": "d", "န": "n", "ပ": "p", "ဖ": "P", "ဗ": "b", "ဘ": "B", "မ": "m",
            "ယ": "y", "ဝ": "w", "သ": "j", "ဟ": "h", "အ": "a", 'လ': 'l', "ဠ": "lx", "ဣ": "ix", "ဤ": "Ix", "်":"F"
        }

            burmese_to_roman.update({
            "၊": "/", "။": "//", "ဥ": "Ux", "ဦ": "OO", "ဧ": "ax", "ဩ": "O", "ဪ": "OR", "ါ": "A", "ာ": "A", "ိ": "i", "ီ": "I","ေ": "e",
            "ု": "u", "ူ": "U", "ဲ": "L", "ံ": "N", "့": ".", "း": ":", "ျ": "Y", "ြ": "R", "ွ": "W", "ှ": "H","၎":"4",
            "ဿ": "jx"
        })
            roman_to_burmese = {v: k for k, v in burmese_to_roman.items()}

            try:
                def roman_to_special_words(text):
                    reverse_rules = [
                        (re.compile(r"q'm"), 'ကျွန် မ '),
                        (re.compile(r"q't"), 'ကျွန် တော် '),
                        (re.compile(r'Q"'), 'ကျွန်ုပ် '),
                    ]
                    for rule in reverse_rules:
                        text = rule[0].sub(rule[1], text)
                    return text

                def romanize_burmese(text):
                    segments = re.split(r"(\{.*?\})", text)  # Splits and keeps braces content as separate segments
                    processed_segments = []

                    for segment in segments:
                        if segment.startswith("{") and segment.endswith("}"):
                            processed_segments.append(segment)
                        else:
                            romanized_text = segment
                            for burmese_char, roman_char in sorted(roman_to_burmese.items(), key=lambda x: len(x[0]), reverse=True):
                                romanized_text = romanized_text.replace(burmese_char, roman_char)
                            processed_segments.append(romanized_text)

                    return "".join(processed_segments)

                transformed_text = ""
                burmese_text = re.sub(r"(,)",r'\1 ',burmese_text)
                burmese_text = roman_to_special_words(burmese_text)
                burmese_text = romanize_burmese(burmese_text)
                pasint = re.sub(r'([က-အ])([က-အ])([က-အ])',r"\1\2္\3", burmese_text)

                for word in pasint.split(" "):
                    word = re.sub(r"([ခဂငဒဝပ]ေ*)ာ", r"\1ါ", word) # If we see the ခဂငဒဝပ the yay chr will be ာ not this ါ
                    word = re.sub(r"([က-အ])(.*)([က-အ])", r"\1\2\3်", word) # if the vowel is followed by the vowel, the a_thet will be applied
                    word = re.sub(r"််", "်", word) # if we see the ််  we remove one
                    word = re.sub(r"်(?=[A-Z])","",word)
                    transformed_text += word + " "
                    transformed_text = re.sub(r"၎ငး", "၎င်း", transformed_text)
                    transformed_text = re.sub(r"([က-အ])္([က-အ])်", r"\1္\2", transformed_text)
                    remove_curly_bracket = transformed_text.replace('{', '').replace('}', '')
                return re.sub(r",\s*",r"",remove_curly_bracket)

            except Exception as e:
                return f"An error occurred during Romanization_to_Burmese: {e}"

# Example usage of the class
if __name__ == "__main__":
    converter = BurmeseConverter()

    # Example: Zawgyi to Unicode
    zawgyi_text = "သီဟိုဠ္မွ ဉာဏ္ႀကီးရွင္သည္ အာယုဝၯနေဆးၫႊန္းစာကို ဇလြန္ေဈးေဘး ဗာဒံပင္ထက္ အဓိ႒ာန္လ်က္ ဂဃနဏဖတ္ခဲ့သည္။"
    try:
        unicode_output = converter.zawgyi_to_unicode(zawgyi_text)
        print("Unicode Output:", unicode_output)
    except Exception as e:
        print(f"Error in Zawgyi to Unicode conversion: {e}")

    # Example: Burmese Romanization
    burmese_text = "လေကြောင်းလိုင်း ခရီးစဉ်အမှတ် ၂၂၈၃ သည် ဆော်ပိုလိုမြို့ လူနေရပ်ကွက်ပေါ်သို့ ပျက်ကျခဲ့ပြီး လိုက်ပါလာသူ ၆၂ ဦးစလုံး သေဆုံးခဲ့သည်။"
    try:
        romanized_output = converter.burmese_to_romanization(burmese_text)
        print("Romanized Output:", romanized_output)
    except Exception as e:
        print(f"Error in Burmese Romanization: {e}")
    
    # Example: Romanization Burmese
    burmese_text = "le kReAc: liuc:, K rI: sq a mHt, ၂ ၂ ၈ ၃, jQ, SeAF piu liu mRiu., lU ne rp kWk peAF jiu., pYk kY KL. pRI:, liuk pA lA jU, ၆ ၂, OO: s luN:, je SuN: KL. jQ // "
    try:
        burmese_output = converter.romanization_to_burmese(burmese_text)
        print("Burmese Output:", burmese_output)
    except Exception as e:
        print(f"Error in Romanization Burmese: {e}")